//==============================================================================
//  EventGeneratorSettings.cpp
//
//  Copyright (C) 2010-2013 Tobias Toll and Thomas Ullrich 
//
//  This file is part of Sartre version: 1.1 
//
//  This program is free software: you can redistribute it and/or modify 
//  it under the terms of the GNU General Public License as published by 
//  the Free Software Foundation.   
//  This program is distributed in the hope that it will be useful, 
//  but without any warranty; without even the implied warranty of 
//  merchantability or fitness for a particular purpose. See the 
//  GNU General Public License for more details. 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//  Author: Thomas Ullrich
//  Last update: 
//  $Date: 2013-05-29 16:26:19 -0400 (Wed, 29 May 2013) $
//  $Author: thomas.ullrich@bnl.gov $
//==============================================================================
#include "EventGeneratorSettings.h"    
#include "Constants.h"    
#include <cmath>    
#include <ctime>    

EventGeneratorSettings* EventGeneratorSettings::mInstance = 0;  // initialize static    

EventGeneratorSettings* EventGeneratorSettings::instance()    
{    
    if (mInstance == 0)     
        mInstance = new EventGeneratorSettings;    
    return mInstance;    
}    

EventGeneratorSettings::EventGeneratorSettings()    
{    
    //   
    //  Register all parameters.   
    //  list() will print them in the order they are defined here.   
    //   
    registerParameter(&mElectronBeamEnergy, "eBeamEnergy", 10.);      
    registerParameter(&mHadronBeamEnergy, "hBeamEnergy", 100.);      
    registerParameter(&mA, "A", static_cast<unsigned int>(1));      
    
    registerParameter(&mNumberOfEvents, "numberOfEvents", static_cast<unsigned long>(10000));    
    registerParameter(&mTimesToShow, "timesToShow", static_cast<unsigned int>(100));    
    
    registerParameter(&mVectorMesonId, "vectorMesonId", 443);  // J/psi    
    registerParameter(&mDipoleModelName, "dipoleModel", string("bSat"));      
    
    registerParameter(&mQ2min, "Q2min", 1000.); // no limits if max <= min    
    registerParameter(&mQ2max, "Q2max", 0.);    
    registerParameter(&mWmin, "Wmin", 1000.);    
    registerParameter(&mWmax, "Wmax", 0.);    
    
    registerParameter(&mCorrectForRealAmplitude, "correctForRealAmplitude", true);    
    registerParameter(&mCorrectSkewedness, "correctSkewedness", true);    
    registerParameter(&mEnableNuclearBreakup, "enableNuclearBreakup", false);    
    registerParameter(&mMaxLambdaUsedInCorrections, "maxLambdaUsedInCorrections", 0.65);    
    registerParameter(&mMaxNuclearExcitationEnergy, "maxNuclearExcitationEnergy", 1.4);    
    registerParameter(&mApplyPhotonFlux, "applyPhotonFlux", true);    
    
    registerParameter(&mVerbose, "verbose", false);    
    registerParameter(&mVerboseLevel, "verboseLevel", 0);    

    registerParameter(&mUPC, "UPC", false);
    registerParameter(&mUPCA, "UPCA", static_cast<unsigned int>(1));
    
    registerParameter(&mRootfile, "rootfile", string("sartre.root"));    
    registerParameter(&mSeed, "seed", static_cast<unsigned int>(time(0)));    
}    

void EventGeneratorSettings::consolidateSettings() // called after runcard is read    
{    
    //    
    //  Check if verbose levels and flags are consistent    
    //  The verbose flag superseeds the verboseLevel.    
    //    
    if (mVerbose && mVerboseLevel == 0) mVerboseLevel = 1;    
    if (mVerboseLevel != 0 && !mVerbose) mVerboseLevel = 0;    
    
    //    
    //  Set random generator seed    
    //    
    mRandomGenerator.SetSeed(mSeed);    
    gRandom->SetSeed(mSeed); // needed for TH1::GetRandom()    
    
    //    
    // Dipole Model    
    //    
    if (mDipoleModelName == string("bSat"))    
        mDipoleModelType = bSat;    
    else if (mDipoleModelName == string("bNonSat"))    
        mDipoleModelType = bNonSat;    
    else     
        mDipoleModelType = bCGC;    
}    

//    
//   Access functions    
//    

void EventGeneratorSettings::setVerbose(bool val) {    
    mVerbose = val;    
    if (mVerbose && mVerboseLevel == 0) mVerboseLevel = 1;    
    if (!mVerbose && mVerboseLevel != 0) mVerboseLevel = 0;    
}    
bool EventGeneratorSettings::verbose() const {return mVerbose;}    

void EventGeneratorSettings::setVerboseLevel(int val) {    
    mVerboseLevel = val;    
    if (!mVerbose && mVerboseLevel != 0) mVerbose = true;    
    if (mVerbose && mVerboseLevel == 0) mVerbose = false;    
}    
int EventGeneratorSettings::verboseLevel() const {return mVerboseLevel;}    

void EventGeneratorSettings::setNumberOfEvents(unsigned long val) { mNumberOfEvents = val;}    
unsigned long EventGeneratorSettings::numberOfEvents() const {return mNumberOfEvents;}    

void EventGeneratorSettings::setTimesToShow(unsigned int val) { mTimesToShow = val;}    
unsigned int EventGeneratorSettings::timesToShow() const {return mTimesToShow;}    

void EventGeneratorSettings::setQ2min(double val) { mQ2min = val;}    
double EventGeneratorSettings::Q2min() const {return mQ2min;}    
double EventGeneratorSettings::Qmin() const {return sqrt(mQ2min);}    

void EventGeneratorSettings::setQ2max(double val) { mQ2max = val;}    
double EventGeneratorSettings::Q2max() const {return mQ2max;}    
double EventGeneratorSettings::Qmax() const {return sqrt(mQ2max);}    

void EventGeneratorSettings::setW2min(double val) { mWmin = sqrt(val);}    
void EventGeneratorSettings::setWmin(double val) { mWmin = val;}    
double EventGeneratorSettings::Wmin() const {return mWmin;}    
double EventGeneratorSettings::W2min() const {return mWmin*mWmin;}    

void EventGeneratorSettings::setW2max(double val) { mWmax = sqrt(val);}    
void EventGeneratorSettings::setWmax(double val) { mWmax = val;}    
double EventGeneratorSettings::Wmax() const {return mWmax;}    
double EventGeneratorSettings::W2max() const {return mWmax*mWmax;}    

int EventGeneratorSettings::vectorMesonId() const {return mVectorMesonId;}    
void EventGeneratorSettings::setVectorMesonId(int val) {mVectorMesonId = val;}    

double EventGeneratorSettings::electronBeamEnergy() const {return mElectronBeamEnergy;}    
void EventGeneratorSettings::setElectronBeamEnergy(double val) {mElectronBeamEnergy = val;}    

double EventGeneratorSettings::hadronBeamEnergy() const {return mHadronBeamEnergy;}    
void EventGeneratorSettings::setHadronBeamEnergy(double val) {mHadronBeamEnergy = val;}    

TLorentzVector EventGeneratorSettings::eBeam() const    
{    
    double mass2;
    if(mUPC)
      mass2=protonMass2;
    else
      mass2=electronMass2;
    return TLorentzVector(0, 0, -sqrt(mElectronBeamEnergy*mElectronBeamEnergy-mass2), mElectronBeamEnergy);  
}    

TLorentzVector EventGeneratorSettings::hBeam() const    
{    
    return TLorentzVector(0, 0, sqrt(mHadronBeamEnergy*mHadronBeamEnergy-protonMass2), mHadronBeamEnergy);    
}    

string EventGeneratorSettings::dipoleModelName() const {return mDipoleModelName;}    
DipoleModelType EventGeneratorSettings::dipoleModelType() const {return mDipoleModelType;}    
void EventGeneratorSettings::setDipoleModelType(DipoleModelType val)   
{  
    mDipoleModelType = val;  
    if (mDipoleModelType == bSat)  
        mDipoleModelName = string("bSat");  
    else if (mDipoleModelType == bNonSat)  
        mDipoleModelName = string("bNonSat");  
    else if (mDipoleModelType == bCGC)  
        mDipoleModelName = string("bCGC");      
}    

string EventGeneratorSettings::rootfile() const {return mRootfile;}    
void EventGeneratorSettings::setRootfile(const char* val) {mRootfile = string(val);}    

unsigned int EventGeneratorSettings::A() const {return mA;}    
void EventGeneratorSettings::setA(unsigned int val) {mA = val;}    

bool EventGeneratorSettings::correctForRealAmplitude() const {return mCorrectForRealAmplitude;}    
void EventGeneratorSettings::setCorrectForRealAmplitude(bool val) {mCorrectForRealAmplitude = val;}    

bool EventGeneratorSettings::correctSkewedness() const {return mCorrectSkewedness;}    
void EventGeneratorSettings::setCorrectSkewedness(bool val) {mCorrectSkewedness = val;}    

bool EventGeneratorSettings::enableNuclearBreakup() const {return mEnableNuclearBreakup;}    
void EventGeneratorSettings::setEnableNuclearBreakup(bool val) {mEnableNuclearBreakup = val;}    

double EventGeneratorSettings::maxNuclearExcitationEnergy() const {return mMaxNuclearExcitationEnergy;}    
void EventGeneratorSettings::setMaxNuclearExcitationEnergy(double val) {mMaxNuclearExcitationEnergy = val;}    

double EventGeneratorSettings::maxLambdaUsedInCorrections() const {return mMaxLambdaUsedInCorrections;}    
void EventGeneratorSettings::setMaxLambdaUsedInCorrections(double val) {mMaxLambdaUsedInCorrections = val;}    

bool EventGeneratorSettings::applyPhotonFlux() const {return mApplyPhotonFlux;}         
void EventGeneratorSettings::setApplyPhotonFlux(bool val) {mApplyPhotonFlux   = val;}       

void EventGeneratorSettings::setUPC(bool val){ mUPC = val; }
bool EventGeneratorSettings::UPC() const { return mUPC; }  

void EventGeneratorSettings::setUPCA(unsigned int val){ mUPCA = val; }
unsigned int EventGeneratorSettings::UPCA() const { return mUPCA; }

