//==============================================================================
//  Kinematics.cpp
//
//  Copyright (C) 2010-2013 Tobias Toll and Thomas Ullrich 
//
//  This file is part of Sartre version: 1.1 
//
//  This program is free software: you can redistribute it and/or modify 
//  it under the terms of the GNU General Public License as published by 
//  the Free Software Foundation.   
//  This program is distributed in the hope that it will be useful, 
//  but without any warranty; without even the implied warranty of 
//  merchantability or fitness for a particular purpose. See the 
//  GNU General Public License for more details. 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//  Author: Thomas Ullrich
//  Last update: 
//  $Date: 2015-05-29 13:37:30 -0400 (Fri, 29 May 2015) $
//  $Author: tobilibob@gmail.com $
//==============================================================================
#include "Kinematics.h"    
#include <cmath>    
#include <iostream>    
    
#define PR(x) cout << #x << " = " << (x) << endl;    
    
using namespace std;    
    
bool Kinematics::mError = false;    
    
bool Kinematics::error() {return mError;}    
    
double Kinematics::xprobe(double Q2, double W2, double vmMass)    
{    
    mError = false;    
    double val = ( Q2 + vmMass*vmMass ) / (W2 - protonMass2 + Q2);    
    if (val > 1 || val < 0) mError = true;    
    return val;    
}       
    
TLorentzVector Kinematics::electronBeam(double eE)    
{    
    mError = false;    
    return TLorentzVector(0, 0, -sqrt(eE*eE-electronMass2), eE);    
}    
    
TLorentzVector Kinematics::hadronBeam(double hE)    
{    
    mError = false;    
    return TLorentzVector(0, 0, sqrt(hE*hE-protonMass2), hE);    
}    
    
double Kinematics::s(const TLorentzVector& e, const TLorentzVector& p)    
{    
    mError = false;    
    return (e+p)*(e+p);    
}    
    
double Kinematics::s(double eE, double hE)    
{    
    mError = false;    
    TLorentzVector e = electronBeam(eE);    
    TLorentzVector p = hadronBeam(hE);    
    return (e+p)*(e+p);    
}    
    
double Kinematics::x(double Q2, double W2)    
{    
    mError = false;    
    double val = Q2/(W2-protonMass2+Q2);    
    if (val > 1 || val < 0) mError = true;    
    return val;    
        
}    
    
double Kinematics::y(double Q2, double x, double s)    
{    
    mError = false;    
    double val = Q2/(x*(s-electronMass2-protonMass2));    
    if (val > 1 || val < 0) mError = true;    
    return val;    
}    
    
double Kinematics::ymin(double s, double vmMass)    
{    
    mError = false;    
    double W2min = Kinematics::W2min(vmMass);    
    double val = (s + W2min - sqrt((s-W2min)*(s-W2min)-4*electronMass2*W2min))/2/(s+electronMass2);    
    if (val > 1 || val < 0) mError = true;    
    return val;    
}    
    
double Kinematics::ymax(double s, double vmMass)    
{    
    mError = false;    
    double W2min = Kinematics::W2min(vmMass);    
    double val = (s + W2min + sqrt((s-W2min)*(s-W2min)-4*electronMass2*W2min))/2/(s+electronMass2);    
    if (val > 1 || val < 0) mError = true;    
    return val;    
}    
    
double Kinematics::W2(double Q2, double x)    
{    
    //    
    //  Returns W2 at a given Q2 and Bjorken x    
    //    
    mError = false;    
    return protonMass2 + Q2*(1-x)/x;    
}    
    
double Kinematics::W2(double Q2, double xprobe, double vmMass)    
{    
    //    
    //  Returns W2 at a given Q2 and xprobe    
    //  where xprobe is     
    //  xprobe = xBJ * (1 + MV^2/Q2)    
    //    
    mError = false;    
    return protonMass2 + vmMass*vmMass/xprobe + Q2*(1-xprobe)/xprobe;    
}    
    
double Kinematics::W2min(double vmMass)    
{    
    mError = false;    
    return vmMass*vmMass + protonMass2;    
}    
    
double Kinematics::W2max(double s)    
{    
    mError = false;    
    return s;   // check !?    
}    
    
double Kinematics::Q2min(double y)    
{    
    mError = false;    
    return electronMass2*y*y/(1-y);    
}    
    
double Kinematics::Q2max(double s)    
{    
    mError = false;    
    return s-electronMass2-protonMass2;    
}    
    
double Kinematics::xpomeron(double t, double Q2, double W2, double vmM)    
{    
    mError = false;    
    double val = (vmM*vmM + Q2 - t)/(W2 + Q2 - protonMass2);  // full expression  
    if (val > 1 || val < 0) mError = true;    
    return val;        
}    

double Kinematics::tmax(double xpom)    
{    
    mError = false;    
    return (- xpom*xpom * protonMass2 / (1-xpom));  // only if m_p (in) = m_p' (out) - to be fixed  
}    
    
double Kinematics::tmax(double t, double Q2, double W2, double vmM)    
{    
    return tmax(xpomeron(t, Q2, W2, vmM));    
}    
    
double Kinematics::tmin(double eE)    
{    
    mError = false;    
    return -2*(eE*protonMass-protonMass2);  // check ?!    
}    
    
bool Kinematics::valid(double s, double t, double Q2, double W2, double vmMass,     
                       bool useTrueXp, bool verbose)    
{    
    //    
    //  Here we perform all complete test of all kinematic variables    
    //    
    double x = Kinematics::x(Q2, W2);    
    double y = Kinematics::y(Q2, x, s);    
        
    //    
    // W    
    //    
    if (W2 < Kinematics::W2min(vmMass)) {    
        if (verbose) cout << "Kinematics::valid(): reject W=" << sqrt(W2) << " because W < Wmin" << endl;    
        return false;    
    }    
    if (W2 > Kinematics::W2max(s)) {    
        if (verbose) cout << "Kinematics::valid(): reject W=" << sqrt(W2) << " because W > Wmax" << endl;    
        return false;    
    }    
    
    //    
    //  y range    
    //    
    double ymin = Kinematics::ymin(s, vmMass);    
    double ymax = Kinematics::ymax(s, vmMass);    
    if (y < ymin || y > ymax) {    
        if (verbose) cout << "Kinematics::valid(): reject y=" << y << " because y < ymin (" << ymin     
                          << ") || y > ymax (" << ymax << ")" << endl;    
        return false;    
    }    
        
    //    
    //  x and y    
    //    
    if (x < 0 || x > 1) {    
        if (verbose) cout << "Kinematics::valid(): reject x=" << x << " because x < 0 || x > 1" << endl;    
        return false;    
    }    
        
    //    
    //  Q2    
    //      
    if (Q2 > Kinematics::Q2max(s)) {    
        if (verbose) cout << "Kinematics::valid(): reject Q2=" << Q2 << " because Q2 > Q2max" << endl;    
        return false;    
    }    
    if (Q2 < Kinematics::Q2min(y)) {    
        if (verbose) cout << "Kinematics::valid(): reject Q2=" << Q2 << " because Q2 < Q2min" << endl;    
        return false;    
    }    
        
    //    
    //  t    
    //    
    //  On calculating xpomeron. xpomeron() provides the correct value  
    //  (not neglecting t or nucleon mass). However, during initialization  
    //  to avoid recursion we set t=0. Calling valid() with t != 0 would  
    //  reject these values. In this case useTrueXp is set to false.  
    //  For final checks if an event has the correct kinematics we have  
    //  to set useTrueXp = true.  
    //  
    double xpom;  
    if (useTrueXp)   
     xpom = Kinematics::xpomeron(t, Q2, W2, vmMass);  
    else  
        xpom = Kinematics::xpomeron(0, Q2, W2, vmMass);  
    double tmax = Kinematics::tmax(xpom);    
    if (t > tmax) {    
        if (verbose) cout << "Kinematics::valid(): reject t=" << t << " because t > tmax (" << tmax << ")" << endl;    
        return false;    
    }    
  
      
    //    
    //  xpom    
    //    
    if (xpom < 0 || xpom > 1) {    
        if (verbose) cout << "Kinematics::valid(): reject xpom=" << xpom << " because xpom < 0 || xpom > 1" << endl;    
        return false;    
    }    
    if (x > xpom) {    
        if (verbose) cout << "Kinematics::valid(): reject since x=" << x << " > xpom = " << xpom << endl;    
        return false;    
    }    
      
    return true;  // survived all tests    
}    
    
