//==============================================================================
//  PhotonFlux.cpp
//
//  Copyright (C) 2010-2013 Tobias Toll and Thomas Ullrich 
//
//  This file is part of Sartre version: 1.1
//
//  This program is free software: you can redistribute it and/or modify 
//  it under the terms of the GNU General Public License as published by 
//  the Free Software Foundation.   
//  This program is distributed in the hope that it will be useful, 
//  but without any warranty; without even the implied warranty of 
//  merchantability or fitness for a particular purpose. See the 
//  GNU General Public License for more details. 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//  Author: Thomas Ullrich
//  Last update: 
//  $Date: 2015-05-29 14:50:23 -0400 (Fri, 29 May 2015) $
//  $Author: thomas.ullrich@bnl.gov $
//==============================================================================
#include "PhotonFlux.h"    
#include "Constants.h"    
#include "Kinematics.h"    
#include "EventGeneratorSettings.h"  
#include "Nucleus.h"
#include <cmath>    
#include "TF1.h"
#include "Math/Functor.h"
#include "Math/IntegratorMultiDim.h"

#define PR(x) cout << #x << " = " << (x) << endl;    

PhotonFlux::PhotonFlux() {
    mS = 0;
    EventGeneratorSettings* settings = EventGeneratorSettings::instance();
    if(settings->UPC()){
        mNucleusUPC = new Nucleus(settings->UPCA());
	mNucleus = new Nucleus(settings->A());
    }
}    

PhotonFlux::PhotonFlux(double s) {
    mS = s;
    EventGeneratorSettings* settings = EventGeneratorSettings::instance();
    if (settings->UPC()) {
        mNucleusUPC = new Nucleus(settings->UPCA());
        mNucleus = new Nucleus(settings->A());
	cout<<"Calculating b-dependence factor...";
	calculateUPCImpactParameterFactor();
	cout<<"Factor="<<mUPCImpactParameterFactor<<endl;
    }
}    

void PhotonFlux::setS(double s) {
    mS = s;
    EventGeneratorSettings* settings = EventGeneratorSettings::instance();
    if(settings->UPC()) {
      cout<<"Calculating b-dependence factor..."<<endl;
      calculateUPCImpactParameterFactor();
      cout<<"Factor="<<mUPCImpactParameterFactor<<endl;
    }
}

double PhotonFlux::operator()(double Q2, double W2, GammaPolarization p) const  
{  
    EventGeneratorSettings* settings = EventGeneratorSettings::instance();
    if (!settings->UPC()) {
        if (p == transverse)  
            return fluxTransverse(Q2, W2);  
        else  
            return fluxLongitudinal(Q2, W2);  
    }
    else {
        if (p == transverse)  
            return nuclearPhotonFlux(Q2, W2);         
        else  
            return 0;
    }
}

double PhotonFlux::fluxTransverse(double Q2, double W2) const    
{    
    //    
    //  Transverse photon flux    
    //    
    double y = Kinematics::y(Q2, Kinematics::x(Q2, W2), mS);    
    
    if (y<0 || y>1 || Kinematics::error()) return 0;      
    if (Q2 > Kinematics::Q2max(mS)) return 0;    
    
    double result = alpha_em/(2*M_PI*Q2*mS*y);    
    result *= (1 + (1-y)*(1-y)) - (2*(1-y)*Kinematics::Q2min(y)/Q2);    
    
    return result;    
}    

double PhotonFlux::fluxLongitudinal(double Q2, double W2) const    
{    
    //    
    //  Longitudinal photon flux    
    //    
    double y = Kinematics::y(Q2, Kinematics::x(Q2, W2), mS);    
    
    if (y<0 || y>1 || Kinematics::error()) return 0;    
    if (Q2 > Kinematics::Q2max(mS)) return 0;    
    
    double result = alpha_em/(2*M_PI*Q2*mS*y);    
    result *= 2*(1-y);    
    
    return result;    
}    

double PhotonFlux::nuclearFormFactor(double q2) const
{
    int A=mNucleusUPC->A();
    double result=0;
    if (A>1) {
        double q=sqrt(q2); //GeV
        double R=mNucleusUPC->radius();//fm
        double rho0=mNucleusUPC->rho0();//fm^-3
        R/=hbarc;//GeV^-1
        rho0*=hbarc*hbarc2; //GeV^3
        double aYukawa=0.7/hbarc; //GeV^-1 (correct value?) 
        
        double term1=4*M_PI*rho0/(A*q*q*q); 
        double term2=sin(q*R)-q*R*cos(q*R); 
        double term3=1./(1.+aYukawa*aYukawa*q*q);
        
        result=term1*term2*term3;
    }
    else {
        double term=1+q2/0.71;
        term*=term;
        result=1./term;
    }
    return result;
}

double PhotonFlux::nuclearPhotonFlux(double Q2, double W2) const
{
    //
    //  This formula is taken from eq. (1) in Phys. Rev. Lett. 84, 11, 2330-2333
    //  The problem is that that equation is wrong (it even has the wrong dimension!)
    //
    //   The proton flux in eq. (7) hep-ph/0310223 seems to be wrong by a factor pi,
    //   *or* wrong in the sense that they mean dn/dkt2dk and not dn/d2ktdk
    //
    //   The equation should read:
    //   dn/dEgamdQ2=Z^2*alpha_em*Q2/(Egam*pi)*((F(Q2+Egam2/LorentzGamma2)/(Q2+Egam2/LorentzGamma2))^2
    //
    //   We need dn/dQ2dW2 which we get from:
    //   dn/dQ2dW2=dn/dEgamdQ2*dEgam/dW2
    //
    //   To Check: Pi factors in A and p
    //   Egamma factor in A
    //
    //   First, make a cut-off on Q2, according to M. Diehl it shoule be
    //   Q2min = Mass2*y*y/(1-y)
    //
    //  double y = Kinematics::y(Q2, Kinematics::x(Q2, W2), mS); ! Need proton y !
    //  double Q2min=protonMass2*y*y/(1-y);
    //  if(Q2 < Q2min)
    //    return 0;
    EventGeneratorSettings* settings = EventGeneratorSettings::instance();
    double beamEnergy=settings->electronBeamEnergy();//Energy per nucleon //GeV
    double eGamma=eGammaOfQ2andW2(Q2, W2); //Lab frame
    double eGamma2=eGamma*eGamma; //GeV^2
    double beamMass=mNucleusUPC->atomicMass()/mNucleusUPC->A(); //Mass per nucleon //GeV
    double beamLorentzGamma2=beamEnergy*beamEnergy/(beamMass*beamMass);
    double Z=mNucleusUPC->Z();
    
    double term1=alpha_em*Z*Z*Q2/eGamma; //GeV^1
    double term2=nuclearFormFactor(Q2+eGamma2/beamLorentzGamma2) / 
    (Q2+eGamma2/beamLorentzGamma2); //GeV^-2
    
    double jacobian=derivativeOfEgamWithW2(); //GeV^-1
    
    double result=term1*term2*term2*jacobian; //GeV^(1-2-2-1)=GeV^-4
    if(result<0) {
      if(settings->verboseLevel() >2)
	cout<<"PhotonFlux::nuclearPhotonFlux(): Warning, Negative photon energy!"<<endl;
      result=0;
    }

    //Crudely compensate for multiple interactions with a universal impactparameter factor:
    result*=mUPCImpactParameterFactor;

    return result;
}

double PhotonFlux::eGammaOfQ2andW2(double Q2, double W2) const
{
    //
    // This function calculates the photon energy 
    // N.B. IN THE LAB FRAME!
    //
    double y= Q2/(Kinematics::x(Q2, W2)*(mS-protonMass2-protonMass2)); 
    EventGeneratorSettings* settings = EventGeneratorSettings::instance();
    double ep=settings->hBeam().E();
    double pz=settings->hBeam().Pz();
    double ee=settings->eBeam().E();
    
    double eMass2=settings->eBeam().M2();
    double term1=-1./(ep+pz);
    double term2=pz/(2*ee)*(Q2+eMass2);
    double term3=y*mS/2;
    
    double result = term1*(term2-term3);
    return result;
}

double PhotonFlux::derivativeOfEgamWithW2() const
{
    EventGeneratorSettings* settings = EventGeneratorSettings::instance();
    double ep=settings->hBeam().E();
    double pz=settings->hBeam().Pz();
    //  double ee=settings->eBeam().E();
    
    double term1=1./(ep+pz); //GeV^-1
    double term3=mS/2./(mS-electronMass2-protonMass2); //GeV^0
    
    return term1*term3; //GeV^-1
    
}

void PhotonFlux::calculateTAAlookupTable(){
  double rbhigh=upperIntegrationLimit*(mNucleus->radius()+mNucleusUPC->radius()); 
  mTAA_of_b=new TH1D("mTAA_of_b", "mTAA_of_b", 1000, 0, rbhigh);
  for(unsigned int i=1; i<=1000; i++){
    double b=mTAA_of_b->GetBinCenter(i);
    mTAA_of_b->SetBinContent(i, TAA(b));
  }
}

void PhotonFlux::calculateUPCImpactParameterFactor() {
  // This function calculates the factor by which the photon flux should be reduced
  // for taking into consideration nuclear-nuclear intearctions 
  // (which do not mix with UPC events experimentally)
  // N.B. The impact parameter here is different from the rest of Sartre. Here it refers
  // to the relative transverse distance between the nuclei, 
  // NOT between the photon and target nucleus.
  //
  // The result is stored in the parameter mUPCImpactParameterFactor
  //
  
  calculateTAAlookupTable();

  double rbhigh=upperIntegrationLimit*(mNucleus->radius()+mNucleusUPC->radius()); 

  TF1* func = new TF1("func",this, &PhotonFlux::bDependenceForIntegration, 0, rbhigh, 0);

  //  double res = func->Integral(0, rbhigh, 0, 1e-3); //fm2
  double res = func->Integral(0., rbhigh);

  double norm=M_PI*rbhigh*rbhigh; //fm2

  mUPCImpactParameterFactor=res/norm;

}
double PhotonFlux::bDependenceForIntegration(double* x, double*) {
  double b=*x;
  //  double probability=exp(-TAA(b)*sigma_nn(mS));
  double probability=exp(-mTAA_of_b->Interpolate(b)*sigma_nn(mS));
  return 2*M_PI*b*probability; //fm
}

double PhotonFlux::sigma_nn(double s) const{
  //parameters http://ntrs.nasa.gov/archive/nasa/casi.ntrs.nasa.gov/20080014212.pdf table 2  
  //I use the mean value of pp and pn collisions: 
  double Z=35.62; //mb
  double Y1=41.34; //mb
  double Y2=31.67; //mb
  double B=0.308; //mb
  double s1=1.0; //GeV2
  double s0=5.38*5.38; //GeV2
  double eta1=0.458;
  double eta2=0.545;

  double result = Z+B*log(s/s0)*log(s/s0)+Y1*pow(s1/s, eta1)-Y2*pow(s1/s, eta2); //mb = 1e-3b = 1e-1 fm2
  result *= 1e-1; //fm2 (1 barn = 1e2 fm2)
  
  return result/hbarc2; //GeV^-2
}

double PhotonFlux::TAA(double b){


  //Integral over dr^2=2*pi*r*dr for

  double rbhigh=upperIntegrationLimit*(mNucleus->radius()+mNucleusUPC->radius()); 

  double lo[2]={0., 0.}; //r, theta
  double hi[2]={rbhigh, 2*M_PI};
  ROOT::Math::Functor wf(this, &PhotonFlux::TAAForIntegration, 2); 
  ROOT::Math::IntegratorMultiDim ig(ROOT::Math::IntegrationMultiDim::kADAPTIVE);
  ig.SetFunction(wf);
  ig.SetAbsTolerance(0.);
  ig.SetRelTolerance(1e-4);
  mB=b;
  
  return ig.Integral(lo, hi); //GeV^2
}

double PhotonFlux::TAAForIntegration(const double* arg) const{
  double r=arg[0]; //fm
  double theta=arg[1];
  double b=mB; //fm

  double arg2=sqrt(r*r+b*b-2*b*r*cos(theta));
  double T1 = mNucleusUPC->A()==1 ? mNucleusUPC->TofProton(r) : mNucleusUPC->T(r); //GeV^2
  double T2 = mNucleus->A()==1 ? mNucleus->TofProton(arg2) : mNucleus->T(arg2); //GeV^2

  return r/hbarc*T1*T2; //GeV^3
}


