//==============================================================================
//  tableMerger.cpp
//
//  Copyright (C) 2010-2013 Tobias Toll and Thomas Ullrich 
//
//  This file is part of Sartre version: 1.1
//
//  This program is free software: you can redistribute it and/or modify 
//  it under the terms of the GNU General Public License as published by 
//  the Free Software Foundation.   
//  This program is distributed in the hope that it will be useful, 
//  but without any warranty; without even the implied warranty of 
//  merchantability or fitness for a particular purpose. See the 
//  GNU General Public License for more details. 
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//  Author: Thomas Ullrich
//  Last update: 
//  $Date: 2013-05-29 16:26:19 -0400 (Wed, 29 May 2013) $
//  $Author: thomas.ullrich@bnl.gov $
//==============================================================================
//   
//  Utility program to merge tables.   
//     
//  Several checks are performed to ensure the integrity of the merged   
//  tables but in any case the new tables should be checked with   
//  tableInspector.    
//   
//  Usage:   
//           tableMerger [-n] [-0] [-p prefix] file(s) ...   
//            -n          Do not create the new tables but perform all checks   
//            -p prefix   Set prefix for output files   
//            -0          Do NOT copy bins with 0 content (ignore them)   
//                        Use when merging partially filled tables  
//            -m          List missing bins (if any)  
//  
//   The output files are named <prefix><ID>_merged.root   
//   where ID is the unique identifier of the table.   
//   
//   Note it is a good idea to always check with option -n on first.   
//   
//==============================================================================   
#include "Table.h"   
#include <iostream>   
#include <vector>   
#include <sstream>   
#include <set>   
#include <unistd.h>   
#include <limits>   
#include <cmath>   
#include <cstdlib>   
#include "TFile.h"    
#include "TH3F.h"    
#define PR(x) cout << #x << " = " << (x) << endl;    
   
using namespace std;   
   
void usage(const char* prog)   
{   
    cout << "Usage: " << prog << " [-n] [-0] [-p prefix] [-m] file(s) ..." << endl;   
}   
   
int main(int argc, char **argv)   
{   
    //   
    //  Handle command line arguments   
    //   
    if (argc == 1) {   
        usage(argv[0]);   
        return 2;   
    }   
       
    bool listMissingBins = false;   
    bool noOutput = false;   
    bool ignoreZero = false;   
    string prefix;   
    int ch;   
    while ((ch = getopt(argc, argv, "m0np:")) != -1) {   
        switch (ch) {   
            case 'm':   
                listMissingBins = true;  
                break;   
            case '0':   
                ignoreZero = true;  
                break;   
            case 'p':   
                prefix = string(optarg);   
                break;   
            case 'n':   
                noOutput = true;   
                break;   
            case '?':   
            default:   
                usage(argv[0]);   
                return 2;   
                break;   
        }   
    }   
    if (optind == argc) {   
        usage(argv[0]);   
        return 2;   
    }   
       
    vector<string> allFiles;   
    for (int index = optind; index < argc; index++)   
        allFiles.push_back(string(argv[index]));   
       
       
    //   
    //  Open the files and get all tables/histograms to be merged   
    //   
    vector<TH3F*> allHistograms;   
    set<uint64_t> idSet;   
    for (unsigned i=0; i<allFiles.size(); i++)  {   
        TFile *file = TFile::Open(allFiles[i].c_str(),"READ");    
        if (file == 0) {    
            cout << "Error: failed opening file '" << allFiles[i].c_str() << "'." << endl;     
            return 1;    
        }    
           
        TH3F *histo = reinterpret_cast<TH3F*>(file->Get("table"));   
        allHistograms.push_back(histo);    
        uint64_t theID = atoll(histo->GetTitle());   
        idSet.insert(theID);   
        if (histo == 0) {    
            cout << "Error: failed retrieving table from file '" << allFiles[i].c_str() << "'." << endl;     
            return 1;    
        }    
        histo->SetDirectory(0);    
        file->Close();    
    }   
    
    //   
    //  Store histos according to their ID   
    //   
    vector<uint64_t> vectorID(idSet.size());   
    copy(idSet.begin(), idSet.end(), vectorID.begin());   
    vector<vector<TH3F*> > vectorHistos(vectorID.size());   
    for (unsigned int i=0; i<vectorID.size(); i++) {   
        for (unsigned int j=0; j<allHistograms.size(); j++) {   
            uint64_t theID = atoll(allHistograms[j]->GetTitle());   
            if (theID == vectorID[i])   
                vectorHistos[i].push_back(allHistograms[j]);   
        }   
    }   
       
    cout << "Will merge: " << endl;   
    for (unsigned int i=0; i<vectorID.size(); i++)   
        cout << "id = " << vectorID[i] << "  => " << vectorHistos[i].size() << " tables" << endl;   
   
    //   
    //  Merge   
    //   
    for (unsigned int i=0; i<vectorID.size(); i++) {  // loop over IDs  
           
        //   
        //  Check if histos to merge have the same bin size   
        //   
        TH3F *refHisto = vectorHistos[i][0];   
        double wx = refHisto->GetXaxis()->GetBinWidth(1);   
        double wy = refHisto->GetYaxis()->GetBinWidth(1);   
        double wz = refHisto->GetZaxis()->GetBinWidth(1);   
        bool notSameBinWidth = false;   
        for (unsigned int j=1; j<vectorHistos[i].size(); j++) {   
            if (fabs(vectorHistos[i][j]->GetXaxis()->GetBinWidth(1) - wx) > numeric_limits<float>::epsilon() ||   
                fabs(vectorHistos[i][j]->GetYaxis()->GetBinWidth(1) - wy) > numeric_limits<float>::epsilon() ||   
                fabs(vectorHistos[i][j]->GetZaxis()->GetBinWidth(1) - wz) > numeric_limits<float>::epsilon()) {   
                cout << "Error: tables for id = " << vectorID[i] << " have different bin width. Cannot merge." << endl;   
                notSameBinWidth = true;   
                break;   
            }   
        }   
        if (notSameBinWidth) continue;   
           
        //   
        // Range of new histogram   
        //   
        double lowerEdgeX = refHisto->GetXaxis()->GetXmin();   
        double upperEdgeX = refHisto->GetXaxis()->GetXmax();   
        double lowerEdgeY = refHisto->GetYaxis()->GetXmin();   
        double upperEdgeY = refHisto->GetYaxis()->GetXmax();   
        double lowerEdgeZ = refHisto->GetZaxis()->GetXmin();   
        double upperEdgeZ = refHisto->GetZaxis()->GetXmax();   
        for (unsigned int j=1; j<vectorHistos[i].size(); j++) {   
            lowerEdgeX = min(vectorHistos[i][j]->GetXaxis()->GetXmin(), lowerEdgeX);   
            upperEdgeX = max(vectorHistos[i][j]->GetXaxis()->GetXmax(), upperEdgeX);   
            lowerEdgeY = min(vectorHistos[i][j]->GetYaxis()->GetXmin(), lowerEdgeY);   
            upperEdgeY = max(vectorHistos[i][j]->GetYaxis()->GetXmax(), upperEdgeY);   
            lowerEdgeZ = min(vectorHistos[i][j]->GetZaxis()->GetXmin(), lowerEdgeZ);   
            upperEdgeZ = max(vectorHistos[i][j]->GetZaxis()->GetXmax(), upperEdgeZ);   
        }   
   
        //   
        // New binning   
        //   
        int nbinX = static_cast<int>((upperEdgeX-lowerEdgeX)/wx + 0.5);   
        int nbinY = static_cast<int>((upperEdgeY-lowerEdgeY)/wy + 0.5);   
        int nbinZ = static_cast<int>((upperEdgeZ-lowerEdgeZ)/wz + 0.5);   
   
        //   
        // Create new histogram   
        //   
        TH3F *mergedHisto = new TH3F("table", refHisto->GetTitle(),    
                                     nbinX, lowerEdgeX, upperEdgeX,    
                                     nbinY, lowerEdgeY, upperEdgeY,   
                                     nbinZ, lowerEdgeZ, upperEdgeZ);    
           
        //   
        // Fill new histo   
        // Here we also make sure that bin centers between old and new match.   
        //    
        double x, y, z, xx, yy, zz, cellContent;   
        for (unsigned int k=0; k<vectorHistos[i].size(); k++) {   
            int nx = vectorHistos[i][k]->GetNbinsX();   
            int ny = vectorHistos[i][k]->GetNbinsY();   
            int nz = vectorHistos[i][k]->GetNbinsZ();   
            for (int ix = 1; ix <= nx; ix++)  {   
                for (int iy = 1; iy <= ny; iy++) {     
                    for (int iz = 1; iz <= nz; iz++) {     
                        // old table   
                        cellContent = vectorHistos[i][k]->GetBinContent(ix, iy, iz);   
                        if (ignoreZero && cellContent == 0) continue; // ignore 0 content  
                        x = vectorHistos[i][k]->GetXaxis()->GetBinCenter(ix);   
                        y = vectorHistos[i][k]->GetYaxis()->GetBinCenter(iy);   
                        z = vectorHistos[i][k]->GetZaxis()->GetBinCenter(iz);   
                           
                        // new table   
                        int globalBin = mergedHisto->FindBin(x, y, z);   
                        if (globalBin == -1) {   
                            cout << "Error while copying data from old to new table. Cannot find referring bin." << endl;   
                            return 1;   
                        }   
                        int kx, ky, kz;   
                        mergedHisto->GetBinXYZ(globalBin, kx, ky, kz);   
                        xx = mergedHisto->GetXaxis()->GetBinCenter(kx);   
                        yy = mergedHisto->GetYaxis()->GetBinCenter(ky);   
                        zz = mergedHisto->GetZaxis()->GetBinCenter(kz);   
                        if (fabs(xx-x) > numeric_limits<float>::epsilon() ||    
                            fabs(yy-y) > numeric_limits<float>::epsilon() ||    
                            fabs(zz-z) > numeric_limits<float>::epsilon()) {   
                            cout << "Error while copying data from old to new table. Bin center mismatch." << endl;   
                            return 1;   
                        }   
                        mergedHisto->SetBinContent(kx, ky, kz, cellContent);   
                    }   
                }   
            }   
        }   
           
        //   
        //  At this point the merged histo is filled.   
        //  Last check is to look for "holes" in the new table.   
        //   
        int nEmpty = 0;   
        int globalBin = 0;  
        vector<int> missingBins;  
        for (int iz = 1; iz <= nbinZ; iz++)    
            for (int iy = 1; iy <= nbinY; iy++)      
                for (int ix = 1; ix <= nbinX; ix++)  {  
                    if (mergedHisto->GetBinContent(ix, iy, iz) == 0) {  
                        nEmpty++;  
                        if (listMissingBins) missingBins.push_back(globalBin);  
                    }  
                    globalBin++;  
                }  
          
        if (nEmpty) {   
            cout << "Warning: the new merged table has " << nEmpty << " cells that are empty. This might indicate that\n";   
            cout << "         there are not enough tables available to fully defines the full cubic range.\n";   
            cout << "         Merged table (" << mergedHisto->GetTitle() << ") is not usable in Sartre." << endl;   
            if (listMissingBins) {  
                cout << "List of missing/empty bins:" << endl;  
                for (unsigned int k=0; k<missingBins.size(); k++) {  
                    cout << missingBins[k];  
                    if (k<missingBins.size()-1) cout << ", ";  
                    if ((k+1)%10 == 0 || k == missingBins.size()-1) cout << "\n";  
                }  
            }  
        }   
          
        //   
        //  All is set now. Unless told not to (-n) we write the histo(s) to file   
        //   
        if (noOutput) {   
            cout << "Tables for id = " << mergedHisto->GetTitle() << " can be merged";   
            if (nEmpty)   
                cout << " but be aware of the empty cells." << endl;  
            else   
                cout << '.' << endl;  
            delete mergedHisto;   
            mergedHisto = 0;   
        }   
        else {   
            ostringstream filenameStream;   
            filenameStream << mergedHisto->GetTitle() << "_merged.root";   
            string filename = prefix + filenameStream.str();   
            TFile *hfile = new TFile(filename.c_str(),"RECREATE");   
            if (!hfile) {   
                cout << "Error writing merged table. Cannot open file '" << filename.c_str() << "'." << endl;   
            }   
            else {   
                mergedHisto->Write();   
                hfile->Close();   
                cout << "Merged table written to file '" << filename.c_str() << "'." << endl;   
                delete mergedHisto;   
                mergedHisto = 0;                   
            }   
        }   
    }   
    cout << "All tables processed." << endl;   
    return 0;   
}   
